/* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */

#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "sector.h"
#include "libburn.h"
#include "drive.h"
#include "transport.h"
#include "message.h"
#include "crc.h"
#include "debug.h"
#include "init.h"
#include "lec.h"
#include "toc.h"
#include "util.h"
#include "sg.h"
#include "write.h"
#include "options.h"


static int type_to_ctrl(int mode)
{
	int ctrl = 0;

	int data = BURN_MODE2 | BURN_MODE1 | BURN_MODE0;
	if (mode & data) {
		ctrl |= 4;
	} else if (mode & BURN_AUDIO) {
		if (mode & BURN_4CH)
			ctrl |= 8;
		if (mode & BURN_PREEMPHASIS)
			ctrl |= 1;
	} else
		assert(0);

	if (mode & BURN_NOCOPY)
		ctrl |= 2;

	return ctrl;
}

/* only the ctrl byte is set here */
static void type_to_form(int mode,
                         unsigned char *ctladr, int *form)
{
	if (mode & BURN_MODE1) {
		*ctladr = 0x40;
		*form = 0x10;
	} else
		printf("oh no! we don't support mode %d\n", mode);
}

void burn_write_flush(struct burn_write_opts *o)
{
	struct burn_drive *d = o->drive;

	if (d->buffer->bytes && !d->cancel) {
		d->write(d, d->nwa, d->buffer);
		d->nwa += d->buffer->sectors;
	}
	d->sync_cache(d);
}

static void add_cue(unsigned char ctladr, unsigned char tno, unsigned char indx,
                    unsigned char form, unsigned char scms,
                    int lba)
{
	int m, s, f;
	burn_lba_to_msf(lba, &m, &s, &f);
	printf(" %1X  %1X | %02X | %02X | %02X | %02X |",
               (ctladr & 0xf0) >> 4, ctladr  & 0xf, tno, indx, form, scms);
	printf("%02X:%02X:%02X\n", dec_to_bcd(m), dec_to_bcd(s), dec_to_bcd(f));
}
static void burn_create_cue_sheet(struct burn_drive *d, struct burn_session *s)
{
	unsigned char ctladr;
	int i, lba = -150;
	int form;

/* if cd text, lead in should be 0x41, and lead in generated by libburn
   if the toc entry pointer is valid, we should use its data here
*/
	printf("\n");
	printf("ctladr|trno|indx|form|scms|  msf\n");
	printf("------+----+----+----+----+--------\n");
/* lead in */
	type_to_form(s->track[0]->mode, &ctladr, &form);
	add_cue(1, 0, 0, 1, 0, lba);
/* gap */
	add_cue(ctladr | 1, 1, 0, form, 0, lba);
	lba += 150;
/* tracks */
	for (i = 0; i < s->tracks; i++) {
		type_to_form(s->track[i]->mode, &ctladr, &form);
/* pregap? */
/* isrc, catalog */
/* data */
		add_cue(ctladr | 1, i+1, 0, form, 0, lba);
/* postgap? */
		lba += burn_track_get_sectors(s->track[i]);
	}
/* lead out */
	add_cue(1, 0xAA, 1, 1, 0, lba);
}

void burn_create_toc_entries(struct burn_drive *d, struct burn_track **tar, int ntr)
{
	int i, m, s, f, runtime = 0;
	struct burn_toc_entry *e;

	burn_print(1, "toc for %d tracks:\n", ntr);
	d->toc_entries = ntr + 3;
	assert(d->toc_entry == NULL);
	d->toc_entry = malloc(d->toc_entries * sizeof(struct burn_toc_entry));
	e = d->toc_entry;
	memset((void *)e, 0, d->toc_entries * sizeof(struct burn_toc_entry));
	e[0].point = 0xA0;
	if (tar[0]->mode & BURN_AUDIO)
		e[0].control = TOC_CONTROL_AUDIO;
	else
		e[0].control = TOC_CONTROL_DATA;
	e[0].pmin = 1;
	e[0].psec = 0;
	e[0].adr = 1;
#warning a0 psec value should be app selectable
	e[1].point = 0xA1;
	e[1].pmin = ntr;
	e[1].adr = 1;
	if (tar[ntr - 1]->mode & BURN_AUDIO)
		e[1].control = TOC_CONTROL_AUDIO;
	else
		e[1].control = TOC_CONTROL_DATA;
	e[2].point = 0xA2;
	e[2].control = e[1].control;
	e[2].adr = 1;
	for (i = 0; i < ntr; i++) {
		tar[i]->entry = &e[3 + i];
		e[3 + i].point = i + 1;
		burn_lba_to_msf(runtime, &m, &s, &f);
		e[3 + i].pmin = m;
		e[3 + i].psec = s;
		e[3 + i].pframe = f;
		e[3 + i].adr = 1;
		e[3 + i].control = type_to_ctrl(tar[i]->mode);
burn_print(1,"track %d control %d\n", tar[i]->mode, e[3+i].control);
		runtime += burn_track_get_sectors(tar[i]);
	}
	burn_lba_to_msf(runtime, &m, &s, &f);
	e[2].pmin = m;
	e[2].psec = s;
	e[2].pframe = f;
burn_print(1, "run time is %d (%d:%d:%d)\n", runtime, m, s, f);
	for (i = 0; i < d->toc_entries; i++)
		burn_print(1, "point %d (%02d:%02d:%02d)\n"
		       , d->toc_entry[i].point
		       , d->toc_entry[i].pmin
		       , d->toc_entry[i].psec
		       , d->toc_entry[i].pframe);
}

int burn_sector_length(int tracktype)
{
	if (tracktype & BURN_AUDIO)
		return 2352;
	if (tracktype & BURN_MODE1)
		return 2048;
	assert(0);
	return 12345;
}

void burn_write_leadin(struct burn_write_opts *o,
                       struct burn_session *s,
                       int first)
{
	struct burn_drive *d = o->drive;
	int count;
	
	d->busy = BURN_DRIVE_WRITING_LEADIN;

	burn_print(5, first?"    first leadin\n":"    leadin\n");

	burn_create_toc_entries(d, s->track, s->tracks);

	if (first)
		count = 0 - d->alba - 150;
	else
		count = 4500;
	
	d->progress.start_sector = d->alba;
	d->progress.sectors = count;
	d->progress.current_sector = 0;
	
	while (count != 0) {
		sector_toc(o, s->track[0]->mode);
		count--;
		d->progress.current_sector++;
	}
	d->busy = BURN_DRIVE_WRITING;
}

void burn_write_leadout(struct burn_write_opts *o,
                        int first, unsigned char control,
                        int mode)
{
	struct burn_drive *d = o->drive;
	int count;

	d->busy = BURN_DRIVE_WRITING_LEADOUT;
/* XXX pass the lead out mode? */
	d->rlba = -150;
	burn_print(5, first?"    first leadout\n":"    leadout\n");
	if (first)
		count = 6750;
	else
		count = 2250;
	d->progress.start_sector = d->alba;
	d->progress.sectors = count;
	d->progress.current_sector = 0;
	
	while (count != 0) {
		sector_lout(o, control, mode);
		count--;
		d->progress.current_sector++;
	}
	d->busy = BURN_DRIVE_WRITING;
}

void burn_write_session(struct burn_write_opts *o,
                        struct burn_session *s)
{
	struct burn_drive *d = o->drive;
	struct burn_track *prev = NULL, *next = NULL;
	int i;

	d->rlba = 0;
	burn_print(1, "    writing a session\n");
	for (i = 0; i < s->tracks; i++) {
		if (i > 0)
			prev = s->track[i-1];
		if (i + 1 < s->tracks)
			next = s->track[i+1];
		else next = NULL;
				
		burn_write_track(o, s->track[i],
                                 prev, next);
	}
}

void burn_write_track(struct burn_write_opts *o,
                      struct burn_track *t, struct burn_track *pt,
                      struct burn_track *nt)
{
#warning XXX if track is not an even sector size we barf
	struct burn_drive *d = o->drive;
	int i, tmp = 0;
	int sectors;
	d->rlba = 0;

	burn_print(1, "        pre-gap\n");

/* pregap interval 1 */
/* is RAW gonna be a problem? */
	if (pt && (t->mode != pt->mode)) {
		for (i = 0; i < 75; i++)
			sector_pregap(o, t->entry->point, pt->entry->control,
                                      pt->mode);
	}
/* pregap interval 2 */
#warning make the first clause match toc control/mode
	if (t->entry->point == 1)
		for (i = 0; i < 150; i++)
			sector_pregap(o, t->entry->point, t->entry->control,
                                      t->mode);
	else if (pt && (t->mode != pt->mode))
		for (i = 0; i < 150; i++)
			sector_pregap(o, t->entry->point, t->entry->control,
                                       t->mode);
/* user data */
	sectors = burn_track_get_sectors(t);
	
	/* Update progress */
	d->progress.start_sector = d->nwa;
	d->progress.sectors = sectors;
	d->progress.current_sector = 0;
	
burn_print(12, "track is %d sectors long\n", sectors);

	if (!nt)
		tmp = sectors > 150 ? 150 : sectors;
burn_print(12, "cutting short by %d sectors\n", tmp);
	for (i = 0; i < sectors - tmp; i++) {
		sector_data(o, t, 0);
		
		/* update current progress */
		d->progress.current_sector++;
	}
	for (; i < sectors; i++) {
		burn_print(1, "last track, leadout prep\n");
		sector_data(o, t, 1);

		/* update progress */
		d->progress.current_sector++;
	}
/* post gap */
	if (nt && !(t->mode & BURN_AUDIO) &&
            (nt->mode & BURN_AUDIO)) {
		burn_print(1, "        postgap\n");
		for (i = 0; i < 150; i++)
			sector_postgap(o, t->entry->point, t->entry->control,
                                       t->mode);
	}
	i = t->offset;
}

void burn_disc_write_sync(struct burn_write_opts *o,
                          struct burn_disc *disc)
{
	struct burn_drive *d = o->drive;
	struct buffer buf;
	struct burn_track *lt;
	int first = 1, i;

	burn_message_clear_queue();

	burn_print(1, "sync write of %d sessions\n", disc->sessions);
	d->buffer = &buf;
	memset(d->buffer, 0, sizeof(struct buffer));
	
	d->rlba = -150;

	d->toc_temp = 9;

	d->send_write_parameters(d, o);
/* Apparently some drives require this command to be sent, and a few drives
return crap.  so we send the command, then ignore the result.
*/
	d->get_nwa(d);
	d->alba = d->start_lba;
	d->nwa = d->alba;

	/* init progress before showing the state */
	d->progress.session = 0;
	d->progress.sessions = disc->sessions;
	d->progress.track = 0;
	d->progress.tracks = disc->session[0]->tracks;
	/* TODO: handle indices */
	d->progress.index = 0;
	d->progress.indices = disc->session[0]->track[0]->indices;
	/* TODO: handle multissession discs */
	/* XXX: sectors are only set during write track */
	d->progress.start_sector = 0;
	d->progress.sectors = 0;
	d->progress.current_sector = 0;

/*burn_create_cue_sheet(d, disc->session[0]);*/
	d->busy = BURN_DRIVE_WRITING;

	for (i = 0; i < disc->sessions; i++) {
		/* update progress */
		d->progress.session = i;
		d->progress.tracks = disc->session[i]->tracks;
		
		burn_write_leadin(o, disc->session[i], first);
		burn_write_session(o, disc->session[i]);
		lt = disc->session[i]->track[disc->session[i]->tracks-1];
		burn_write_leadout(o, first, lt->entry->control,
                                   lt->mode);
		if (first)
			first = 0;
		
		/* XXX: currently signs an end of session */
		d->progress.current_sector = 0;
		d->progress.start_sector = 0;
		d->progress.sectors = 0;
	}
	burn_write_flush(o);
	sleep(1);

	burn_print(1, "done\n");
	d->busy = BURN_DRIVE_IDLE;
}
