/* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */

#include "writer.h"
#include "util.h"
#include "volume.h"
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#define APPLICATION_ID "LIBBURN SUITE (C) 2002 D.FOREMAN/B.JANSENS"

static void iso_next_state (struct iso_write_target *target);
static int iso_source_get_size (struct burn_source *src);
static void iso_source_free (struct burn_source *src);
static void iso_target_layout (struct iso_write_target *target);
static void iso_dir_layout (struct iso_write_target *target,
			    struct iso_tree_dir **dir);
static int iso_write_system_area (struct iso_write_target *t,
				  unsigned char *buffer,
				  enum burn_source_status *err);
static int iso_write_pri_volume (struct iso_write_target *t,
				 unsigned char *buffer,
				 enum burn_source_status *err);
static int iso_write_volume_terminator (struct iso_write_target *t,
					unsigned char *buffer,
					enum burn_source_status *err);
static int iso_write_path_table (struct iso_write_target *t,
				 unsigned char *buffer,
				 enum burn_source_status *err,
				 int lsb);
static int iso_write_dir_record (struct iso_write_target *target,
				 unsigned char *buffer,
				 struct iso_tree_dir **dir);
static void iso_write_file_id (unsigned char *buf, int size,
			       struct iso_tree_file **f);

struct burn_source*
iso_source_new (struct iso_volumeset *volumeset,
		int volume)
{
	struct burn_source *src;
	struct iso_write_target *t;

	if (!volumeset) return NULL;
	if (volume >= volumeset->volumeset_size) return NULL;

	t = malloc (sizeof (struct iso_write_target));

	iso_tree_sort (volumeset->root);

	t->volset = volumeset;
	t->volset->refcount++;
	t->volume = volume;
	t->now = time (NULL);
	t->phys_sector_size = 2048;
	iso_target_layout (t);

	t->state = ISO_WRITE_BEFORE;
	iso_next_state (t); /* prepare for the first state */

	src = malloc (sizeof (struct burn_source));
	src->refcount = 1;
	src->free_data = iso_source_free;
	src->read = iso_source_generate;
	src->read_sub = NULL;
	src->get_size = iso_source_get_size;
	src->data = t;
	return src;
}

void
iso_source_free (struct burn_source *src)
{
	struct iso_write_target *target = src->data;

	assert (src->read == iso_source_generate &&
		src->read_sub == NULL &&
		src->get_size == iso_source_get_size);

	iso_volumeset_free (target->volset);
	free (target);
}

void
iso_dir_layout (struct iso_write_target *target,
		struct iso_tree_dir **ddir)
{
	int i;
	struct iso_tree_dir *dir = *ddir;

	dir->logical_block = 0;
	dir->size = 0;

	for (i = 0; i < dir->nfiles; ++i) {
		dir->files[i].logical_block = 0;
		dir->files[i].size = 0;
	}

	for (i = 0; i < dir->nchildren; ++i)
		iso_dir_layout (target, dir->children[i].me);
}

void
iso_target_layout (struct iso_write_target *target)
{
	target->total_size = 0;
	/* system area */
	target->total_size += 16 * target->phys_sector_size;
	/* primary volume descriptor */
	target->total_size += target->phys_sector_size;
	/* volume descriptor terminator */
	target->total_size += target->phys_sector_size;

	target->volume_space_size = 0;
	target->logical_block_size = 2048;
	target->path_table_size = 0;
	target->l_path_table_pos = 0;
	target->m_path_table_pos = 0;

	iso_dir_layout (target, target->volset->root);
}

int
iso_source_get_size (struct burn_source *src)
{
	struct iso_write_target *target = src->data;

	assert (src->read == iso_source_generate &&
		src->read_sub == NULL &&
		src->get_size == iso_source_get_size);

	return target->total_size;
}

void
iso_next_state (struct iso_write_target *target)
{
	switch (++target->state)
	{
	case ISO_WRITE_BEFORE:
		/* this should never occur */
		assert (0);
		break;
	case ISO_WRITE_SYSTEM_AREA:
		target->state_data.system_area.sectors = 0;
		break;
	case ISO_WRITE_PRI_VOL_DESC:
		break;
	case ISO_WRITE_VOL_DESC_TERMINATOR:
		break;
	case ISO_WRITE_L_PATH_TABLE:
		break;
	case ISO_WRITE_M_PATH_TABLE:
		break;
	case ISO_WRITE_DONE:
		break;
	}
}

enum burn_source_status
iso_source_generate (struct burn_source *src,
		     unsigned char *buffer,
		     int size)
{
	int next = 0;
	enum burn_source_status err = BURN_SOURCE_OK;
	struct iso_write_target *target = src->data;

	assert (src->read == iso_source_generate &&
		src->read_sub == NULL &&
		src->get_size == iso_source_get_size);


	/* make sure the app didn't fuck up badly. */
	if (size != target->phys_sector_size)
		return BURN_SOURCE_FAILED; /* XXX better code */

	switch (target->state)
	{
	case ISO_WRITE_BEFORE:
		/* this should never occur */
		assert(0);
	case ISO_WRITE_SYSTEM_AREA:
		next = iso_write_system_area (target, buffer, &err);
		break;
	case ISO_WRITE_PRI_VOL_DESC:
		next = iso_write_pri_volume (target, buffer, &err);
		break;
	case ISO_WRITE_VOL_DESC_TERMINATOR:
		next = iso_write_volume_terminator (target, buffer, &err);
		break;
	case ISO_WRITE_L_PATH_TABLE:
		next = iso_write_path_table (target, buffer, &err, 1);
		break;
	case ISO_WRITE_M_PATH_TABLE:
		next = iso_write_path_table (target, buffer, &err, 0);
		break;
	case ISO_WRITE_DONE:
		err = BURN_SOURCE_EOF;
		break;
	}

	if (next)
		iso_next_state (target);

	return err;
}


/* writes 16 sectors of '0' */
int
iso_write_system_area (struct iso_write_target *t,
		       unsigned char *buffer,
		       enum burn_source_status *err)
{
	struct iso_state_system_area *state = &t->state_data.system_area;

	memset (buffer, 0, t->phys_sector_size);
	return (++state->sectors == 16);
}

/* writes the primary volume descriptor */
int
iso_write_pri_volume (struct iso_write_target *t,
		      unsigned char *buffer,
		      enum burn_source_status *err)
{
	/* volume descriptor type (8.4.1) */
	buffer[0] = 1;
	/* standard identifier (8.4.2) */
	memcpy (&buffer[1], "CD001", 5);
	/* volume descriptor version (8.4.3) */
	buffer[6] = 1;
	/* unused field (8.4.4) */
	buffer[7] = 0;
	/* system identifier (8.4.5) */
	/* XXX mkisofs puts "LINUX" here */
	memset (&buffer[8], ' ', 32);
	/* volume identifier (8.4.6) */
	iso_d_strcpy (&buffer[40], 32, t->volset->volume_id[t->volume]);
	/* unused field (8.4.7) */
	memset (&buffer[72], 0, 8);
	/* volume space size (8.4.8) */
	iso_bb (&buffer[80], t->volume_space_size, 4);
	/* unused field (8.4.9) */
	memset (&buffer[88], 0, 32);
	/* volume set size (8.4.10) */
	iso_bb (&buffer[120], t->volset->volumeset_size, 2);
	/* volume sequence number (8.4.11) */
	iso_bb (&buffer[124], t->volume + 1, 2);
	/* logical block size (8.4.12) */
	iso_bb (&buffer[128], t->logical_block_size, 2);
	/* path table size (8.4.13) */
	iso_bb (&buffer[132], t->path_table_size, 4);
	/* location of occurance of type l path table (8.4.14) */
	iso_lsb (&buffer[140], t->l_path_table_pos, 4);
	/* location of optional occurance of type l path table (8.4.15) */
	iso_lsb (&buffer[144], 0, 4);
	/* location of occurance of type m path table (8.4.16) */
	iso_msb (&buffer[148], t->m_path_table_pos, 4);
	/* location of optional occurance of type m path table (8.4.17) */
	iso_msb (&buffer[152], 0, 4);
	/* directory record for root directory (8.4.18) */
	iso_write_dir_record (t, &buffer[156], t->volset->root);
	/* volume set identifier (8.4.19) */
	iso_d_strcpy (&buffer[190], 128, t->volset->volumeset_id);
	/* publisher identifier (8.4.20) */
	iso_a_strcpy (&buffer[318], 128, t->volset->publisher_id);
	/* data preparer identifier (8.4.21) */
	iso_a_strcpy (&buffer[446], 128, t->volset->data_preparer_id);
	/* application identifier (8.4.22) */
	iso_a_strcpy (&buffer[574], 128, APPLICATION_ID);
	/* copyright file identifier (8.4.23) */
	iso_write_file_id (&buffer[702], 37, t->volset->copyright_file);
	/* abstract file identifier (8.4.24) */
	iso_write_file_id (&buffer[739], 37, t->volset->abstract_file);
	/* bibliographic file identifier (8.4.25) */
	iso_write_file_id (&buffer[776], 37, t->volset->bibliographic_file);
	/* volume creation date and time (8.4.26) */
	/* XXX is this different for multisession? */
	iso_datetime_17 (&buffer[813], t->now);
	/* volume modification date and time (8.4.27) */
	iso_datetime_17 (&buffer[830], t->now);
	/* volume expiration date and time (8.4.28) */
	iso_datetime_17 (&buffer[847], t->volset->expiration_time);
	/* volume effective date and time (8.4.29) */
	iso_datetime_17 (&buffer[864],
			 t->volset->effective_time == (time_t) -1 ?
			 t->now : t->volset->effective_time);
	/* file structure version (8.4.30) */
	buffer[881] = 1;
	/* reserved for future standardization (8.4.31) */
	buffer[882] = 0;
	/* application use (8.4.32) */
	/* XXX put something in here? does mkisofs? */
	memset (&buffer[883], 0, 512);
	/* reserved for future standardization (8.4.33) */
	memset (&buffer[1395], 0, 653);

	return 1;
}

void
iso_write_file_id (unsigned char *buf, int size, struct iso_tree_file **f)
{
	if (!f)
		memset (buf, ' ', size);
	else
		iso_filecpy (buf, size,
			     iso_tree_file_get_name (f, ISO_FILE_NAME_ISO),
			     (*f)->version);
}

int
iso_write_dir_record (struct iso_write_target *t,
		      unsigned char *buffer,
		      struct iso_tree_dir **ddir)
{
	int len_fi;
	const char *fi = NULL;
	int sz;
	struct iso_tree_dir *dir = *ddir;

	if (!dir->fullname)
		len_fi = 1;
	else {
		fi = iso_tree_dir_get_name (ddir, ISO_FILE_NAME_ISO);
		len_fi = strlen (fi);
	}

	sz = 33 + len_fi + !(len_fi % 2);

	/* length of directory record (9.1.1) */
	buffer[0] = sz;
	/* extended attribute record length (9.1.2) */
	buffer[1] = 0; /* XXX allow for extended attribute records */
	/* location of extent (9.1.3) */
	iso_bb (&buffer[2], dir->logical_block, 4);
	/* data length (9.1.4) */
	iso_bb (&buffer[10], dir->size, 4);
	/* recording date and time (9.1.5) */
	iso_datetime_7(&buffer[18], t->now);
	/* file flags (9.1.6) */
	buffer[25] = ISO_FILE_FLAG_DIRECTORY;
	/* file unit size (9.1.7) */
	buffer[26] = 0; /* XXX support this shit? */
	/* interleave gap size (9.1.8) */
	buffer[27] = 0; /* XXX support this shit? */
	/* volume sequence number (9.1.9) */
	iso_bb (&buffer[28], dir->volume + 1, 2);
	/* length of file identifier (9.1.10) */
	buffer[32] = len_fi;
	/* file identifier */
	if (!dir->fullname)
		buffer[33] = 0;
	else
		iso_d_strcpy (&buffer[33], len_fi, fi);
	if (!(len_fi % 2))
		buffer[33 + len_fi] = 0;

	return sz;
}

/* writes the volume descriptor set terminator */
int
iso_write_volume_terminator (struct iso_write_target *t,
			     unsigned char *buffer,
			     enum burn_source_status *err)
{
	/* volume descriptor type (8.3.1) */
	buffer[0] = 255;
	/* standard identifier (8.3.2) */
	memcpy (&buffer[1], "CD001", 5);
	/* volume descriptor version (8.3.3) */
	buffer[6] = 1;
	/* reserved for future standardization (8.3.4) */
	memset (&buffer[7], 0, 2041);

	return 1;
}

/* writes set of path table records */
int iso_write_path_table (struct iso_write_target *t,
			  unsigned char *buffer,
			  enum burn_source_status *err,
			  int lsb)
{
	struct iso_state_path_tables *state = &t->state_data.path_tables;


	return 1;	
}
